#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' microcsound, a front end for composition with Csound, with an
emphasis on pitch flexibility/microtonality. '''

## version 2017-02-24. For a list of changes, see the CHANGES file.

##### EDIT DEFAULTS HERE:
##### orchestra dir, where the script can look for your
##### Csound orchestra files:
ORC_DIR = '/usr/local/share/microcsound'
##### the name, without a full path, of the default Csound orchestra
##### file that microcsound will look for if none is explicitly stated
DEFAULT_ORC_FILE = 'microcsound_basic.orc'
##### the beginning stub of a Csound command for non-realtime use:
NORMAL_CSOUND_COMMAND_STUB = '/usr/bin/csound -d --messagelevel=0 ' \
                             '--nodisplays -W'
##### The beginning stub Csound command for realtime use:
RT_CSOUND_COMMAND_STUB = '/usr/bin/csound -d --messagelevel=0 --nodisplays  '\
                         '-b2048 -B2048 -odac'
##### END USER EDITING

MIDDLE_C_HZ = 261.6255653
### for changing middle c numerical pitch standard
### in the notation (MIDI=5, CSound=8, piano=4)
### MIDI is default:
MIDDLE_C_OCTAVE = 5

from readline import *
from sys import stdin, stdout, argv
from os import system
from random import gauss
from math import log
import re
import argparse

######################################

##### this below allows the parser to use a compiled regex pattern to know
##### what to look for in a statement

PARSER_PATTERN = re.compile(
    r"(?:(?:t|i|mix|pan|div|gr|gv|gs)[=](?:[0-9]{0,5}[.]?[0-9]{1,5}|\"<\"))"
    r"|(?:(?:p[89]{1}|p[1-9][0-9])[=](?:[-0-9.<]+))"
    r"|(?:key[=][0-9]+[:][0-9]+)"
    r"|(?:\"[-0-9.<%]+\")"
    r"|(?:[&][\-+]?[0-9]*)"
    r"|(?:\[L:[0-9]{1,4}[/][0-9]{1,4}\])"
    r"|(?:[0-9]{1,4}[/][0-9]{1,4})"
    r"|[rzx][0-9]{0,2}"
    r"|[@](?:[0-9]?[.][0-9]{1,3}|[0-9A-Fa-f]{1,2}|[<])"
    r"|PD[0-9]{1,3}"
    r"|PU"
    r"|(?:(?:[.\(])?(?:\^/2|_/2|[_^=/\\<>!?]|\xc2\xa1|\xc2\xbf)*[a-g](?:\*)*[,']*[0-9]{0,2}[-]?[)]?)"
    r"|(?:(?:[.\(])?(?:[0-9]+[:][0-9]+)\)?(?:[| t](?![=]))*)"
    r"|(?:(?:[.\(])?(?:[0-9]+[.])?(?:[-]?[0-9]+)\)?(?:[| t](?![=]))*)"
    r"|\["
    r"|\]")

class State(object):
    ## storing some state in a safe object interface.
    ## these are the variables that would be annoying to have to redefine
    ## for every voice line...
    def __init__(self):
        self.div = 31
        self.instr = 1
        self.tempo = 120
        self.tempostring = 't 0 120 ' ###
        self.outstring = 'i200 0 -1\n' ### everything output goes here
                            ### added starting zak mixer statement on
                            ### 2010-08-03
                            ### instead of an empty starting string
        self.new_voice()
    
    def new_voice(self):
        self.length = 1/4.
        self.octave = 5
        self.articulation = 'non-legato'
        self.gaussian_rhythm = 0
        self.gaussian_volume = 0 ### gaussian volume variance
        self.gaussian_staccato = 0 ### gaussian staccato variance

state = State()

#### some helper functions:

def octaves(pitch):
    """return a value in log octaves of a pitch expressed as a decimal"""
    return log(pitch, 2)

def solfege2et(text, div):
    ''' translate an traditional "abcdefg" plus accidentals notation
    into numerical form. "div" represents the division of the octave
    we are working with. '''

    ### some helper definitions:
    octav = int(round(div))    
    tempered_fifth = round(octaves(1.5) * div)
    whole_step = 2 * tempered_fifth % div
    chromatic = (3 * whole_step) - \
                ((-1 * tempered_fifth) % div)
    half_chromatic = chromatic * .5
    syntonic = round(octaves(81/80.0) * div)
    septimal = round(octaves(64/63.0) * div)
    undecimal = round(octaves(33/32.0) * div)
    tridecimal = round(octaves(1053/1024.0) * div)
    ### extra: ###
    pure_fifth = octaves(1.5) * div
    meancomma = pure_fifth - tempered_fifth

    ### the notes:
    notes = {'=': 0,
             'c': 0,
             'd': whole_step,
             'e': whole_step * 2,
             'f': (-1 * tempered_fifth) % div,
             'g': tempered_fifth,
             'a': tempered_fifth + whole_step,
             'b': tempered_fifth + whole_step * 2,
             '^': chromatic,
             '_': -1 * chromatic,
             '^/2': half_chromatic,
             '_/2': -1 * half_chromatic,
             '/':  syntonic,
             '\\': -1 * syntonic,
             '>': septimal,
             '<': -1 * septimal,
             '!': undecimal,
             '¡': -1 * undecimal,
             '?': tridecimal,
             '¿': -1 * tridecimal,
             '*': meancomma,
             "'": octav,
             ",": -1 * octav
            }

    # a place to collect the total:
    value_sum = 0

    for i in re.findall(r"\^/2|_/2|[a-g',^_=/\\<>!?]|\xc2\xa1|\xc2\xbf|\*",
                        text):
        value_sum += notes[i]

    ### return the result:
    return int(value_sum)

def degree2hz(degree, div):
    """return a pitch in HZ from a numerical index
    and a division of the octave"""

    myhz = MIDDLE_C_HZ * pow(2, (degree+0.0)/div)
    return "%1.5f" % myhz

def handle_symbolic_notation(event):
    ''' interpret and return data from a symbolic note event '''

    mylist = re.search(r"(?P<articul>[.\(]?)"
                     r"(?P<pitch>(?:\^/2|_/2|[_^=/\\<>!?]|"
                     r"\xc2\xa1|\xc2\xbf)*"
                     r"[a-g](?:\*)*[',]*)"
                     r"(?P<len>[0-9]{0,2})(?P<tie>[-]?)"
                     r"(?P<legato_end>[)]?)",
                     event)
    articul = mylist.group('articul')
    pitch = mylist.group('pitch')
    length_factor = mylist.group('len')
    tie_dash = mylist.group('tie')
    legato_end = mylist.group('legato_end')

    if articul and articul == '(':
        state.articulation = 'legato'
    if legato_end:
        state.articulation = 'non-legato'
    if articul and articul == '.':
        articulation = 'staccato'
    else:
        articulation = state.articulation

    if pitch:
        degree = solfege2et(pitch, state.div)
        pitch = degree2hz(degree, state.div)
    if length_factor:
        length_factor = int(length_factor)
    else:
        length_factor = 1

    if tie_dash:
        tie = 1
    else:
        tie = 0

    return pitch, length_factor, articulation, tie

def handle_numeric_notation(event):
    ''' handle and return data from a numeric notation event '''

    mylist = re.search(r"(?P<articul>[.\(]?)"
             r"(?:(?P<oct>[0-9]+)[.])?"
             r"(?P<deg>[-]?[0-9]+)"
             r"(?P<legato_end>[\)]?)"
             r"(?P<tie>[| t]*)", event)
    articul = mylist.group('articul')
    try:
        myoct = mylist.group('oct')
    except:
        print event
    deg = mylist.group('deg')
    legato_end = mylist.group('legato_end')
    tie_phrase = mylist.group('tie')

    if articul and articul == '(':
        state.articulation = 'legato'
    if legato_end:
        state.articulation = 'non-legato'
    if articul and articul == '.':
        articulation = 'staccato'
    else:
        articulation = state.articulation

    if myoct:
        state.octave = int(myoct)
    if deg:
        degree_raw = int(deg)
        if state.div > 0:
            degree = (state.octave - MIDDLE_C_OCTAVE) * state.div + degree_raw
            pitch = degree2hz(degree, state.div)
        else:
            ### when div=0, pitch is unintepreted
            pitch = degree_raw

    length_factor = 1
    if tie_phrase:
        length_factor += tie_phrase.count('t')

    ### we've already handled the ties ourselves:
    tie = 0

    return pitch, length_factor, articulation, tie

def handle_JI_notation(event):
    ''' handle and return data from a JI note event '''

    #print "aha, a ratio!"
    mylist = re.search(r"(?P<articul>[.\(]?)"
                     r"(?P<ratio>[0-9]+[:][0-9]+)"
                     r"(?P<legato_end>[\)]?)"
                     r"(?P<tie>[| t]*)", event)
    articul = mylist.group('articul')
    try:
        ratio_text = mylist.group('ratio')
    except:
        print "malformed or absent ratio: %s" % event
    legato_end = mylist.group('legato_end')
    tie_phrase = mylist.group('tie')

    if articul and articul == '(':
        state.articulation = 'legato'
    if legato_end:
        state.articulation = 'non-legato'
    if articul and articul == '.':
        articulation = 'staccato'
    else:
        articulation = state.articulation

    ratio_text_new = ratio_text.split(':')
    ratio = float(ratio_text_new[0])/float(ratio_text_new[1])
    pitch = MIDDLE_C_HZ * ratio

    length_factor = 1
    if tie_phrase:
        length_factor += tie_phrase.count('t')

    ## we've already handled the tie:
    tie = 0

    return pitch, length_factor, articulation, tie

######### parsing/playing function, the heart of the program:
def parser(inst_line):

    ### before parsing this voice, set up the starting points:
    
    ## reset SOME of the state variables:
    state.new_voice()

    ## basic variables
    grid_time = 0
    chord_status = 0
    xtra = ''

    ## pitch related defaults:
    octave = MIDDLE_C_OCTAVE
    degree = 0
    key = 1

    length_factor = 1
    tie_dict = {}
    tie_dict[state.instr] = {}
    tie = 0

    staccato_length = .2
    non_legato_space = .02

    pedal_down = False

    default_attack = .66
    pan = 0.5
    mix = 1

    for event in PARSER_PATTERN.findall(inst_line):
        ## a non-event?
        if event == '':
            continue

        ## a variable assignment:
        elif re.match(r"(?P<type>div|mix|pan|gr|gv|gs|t|i)[=]"
                      r"(?P<value>(?:[0-9]{1,5}[.]?[0-9]{0,5}|\"<\"))",
                      event):
            global_variable = event.split('=')
            evtype, val = global_variable[0], global_variable[1]
            if evtype == 't':
                state.tempo = float(val)
                if grid_time == 0:
                    state.tempostring = 't %s %s ' % (grid_time, val)
                else:
                    state.tempostring = state.tempostring + '%s %s ' \
                                                          % (grid_time, val)
            elif evtype == 'div':
                state.div = float(val)
            elif evtype == 'i':
                state.instr = float(val)
                tie_dict[state.instr] = {}
            elif evtype == 'pan':
                if val == '<':
                    val = "<"
                pan = val
            elif evtype == 'gr':
                state.gaussian_rhythm = float(val)
            elif evtype == 'gv':
                state.gaussian_volume = float(val)
            elif evtype == 'gs':
                state.gaussian_staccato = float(val)
            else:
                if val == '<':
                    val = "<"
                mix = val
            continue
       
        ## modify a pre-existing instrument parameter slot. Parameters slots
        ## MUST have been pre-declared, or an exception is thrown.
        elif re.match(r"(?:(?:p[89]{1}|p[1-9][0-9])[=](?:[-0-9.<]+))",
                      event):
            xtext = event.split('=')
            pslot = int(xtext[0].replace('p', '')) - 8
            try:
                xtra_as_array = xtra.split(' ')
                xtra_as_array[pslot] = xtext[1]
                xtra = ' '.join(xtra_as_array)
            except:
                print "Instrument parameters must be declared as a string "\
                      "before they can be altered! Problem statement is: %s" \
                      % event
                exit()
            continue
        
        ## transposition by JI ratio:
        elif 'key' in event:
            key_ratio_text = event.split('=')[1]
            key_ratio_text_new = key_ratio_text.split(':')
            key = float(key_ratio_text_new[0])/float(key_ratio_text_new[1])
            continue

        ## xtra instrument parameters:
        elif re.match(r"(?:\"[-0-9.<%]+\")", event):
            xtra = event.replace('%', ' ').replace('"','')
            continue

        ## lengths:
        elif re.match(r"(?:\[L:)?[0-9]{1,4}[/][0-9]{1,4}(?:\])?", event):
            z = re.search(r"(?:\[L:)?([0-9]{1,4})[/]([0-9]{1,4})(?:\])?",
                          event)
            numerator = float(z.group(1))
            denominator = float(z.group(2))
            state.length = (numerator/denominator) * 4
            continue

        ## rests:
        elif event[0] == 'r' or event[0] == 'z' or event[0] == 'x':
            factor = event[1:]
            if factor != '':
                length_factor = int(factor)
            else:
                length_factor = 1
            grid_time += (state.length * length_factor)
            continue

        ## chords are enabled by 'stopping the clock':
        elif event == '[':
            chord_status = 1
            continue
        elif event == ']':
            chord_status = 0
            grid_time += (state.length * length_factor)
            continue

        ## added this for sustain pedal passages:
        elif 'PD' in event:
            ev = re.match(r'PD(?P<arrival>\d{1,3})', event)
            pedal_down = True
            arrival = (int(ev.group('arrival')) * state.length) + grid_time
            continue
        elif 'PU' in event:
            pedal_down = False
            continue

        ## grid time pointer can be changed for 'time travel' :-) :
        elif event[0] == '&':
            r = re.search(r"[&]([\-+]?)([0-9]*)", event)
            sign = r.group(1)
            value = r.group(2)
            if sign:
                if sign == '-':
                    grid_time -= state.length * int(value)
                if sign == '+':
                    grid_time += state.length * int(value)
            else:
                grid_time = state.length * int(value)
            continue

        ## attack level:
        elif event[0] == '@':
            if '.' in event[1:]:
                default_attack = float(event[1:])
            elif '<' in event[1:]:
                default_attack = '<'
            else:
                if len(event[1:]) == 2:
                    default_attack = eval('%s/99.' % event[1:])
                elif len(event[1:]) == 1:
                    default_attack = eval('%s0/99.' % event[1:])
            continue

        ####################
        ## pitch events : ##
        ####################

        ## ratio notation (JI):
        elif re.match(r"(?:[.\(])?(?:[0-9]+[:][0-9]+)", event):
            pitch, length_factor, articulation, tie = \
                handle_JI_notation(event)

        ## [oct.]degree notation:
        elif re.match(r"(?:[.\(])?(?:[0-9]+[.])?(?:[-]?[0-9]+)", event):
            pitch, length_factor, articulation, tie = \
                handle_numeric_notation(event)

        ## symbolic diatonic notation:
        else:
            pitch, length_factor, articulation, tie = \
                handle_symbolic_notation(event)

        ## possibly we have a JI transposition:
        pitch = pitch * key

        ### OK, if we have finally gotten a pitch event, we know what's what
        ### now. If we've gotten this far in the loop,
        ### we can calculate stuff....
        lf = length_factor
        on_time = grid_time + (gauss(0, 0.001) * state.gaussian_rhythm * \
                               state.tempo * 0.01666)
        if on_time < 0:
            on_time = 0
        if default_attack != '<':
            attack = default_attack + int(gauss(0, state.gaussian_volume))
            if attack >= 1:
                attack = 1
        else:
            attack = default_attack

        ### this next block take into account whether the not is a tie or a
        ### regular note event, and takes appropriate actions:
        if tie:
            ### if note is not in the list of tied notes, add it.
            if pitch not in tie_dict[state.instr]:
                tie_dict[state.instr][pitch] = [on_time, length_factor,
                                           attack, pan]
            else:                      ### if it IS there already,
                                       ### add duration to it.
                tie_dict[state.instr][pitch][1] = tie_dict[state.instr][degree][1] + \
                                             length_factor
            continue
        if pitch in tie_dict[state.instr]:
            ### or if the note was a tie, and is now not tied, that means
            ### it's time actually get its endpoints and send it to the
            ### output list:
            event = tie_dict[state.instr].pop(pitch)
            on_time = event[0]
            lf = event[1] + length_factor
            attack = event[2]
            pan = event[3]

        ### calculate duration:
        if pedal_down:
            duration = arrival - on_time
        else:
            if articulation == 'staccato':
                duration = ((state.tempo/60.) * staccato_length) + \
                           (gauss(0, 0.001) * state.gaussian_staccato * \
                           state.tempo * 0.01666)
            elif articulation == 'legato':
                duration = (state.length * lf) * -1  ### negative p3 for
                                               ### legato instruments
            else:
                ### non-legato (state.tempo * 0.01666?)
                duration = (state.length * lf) # - \
                           #(state.tempo * non_legato_space / 60.) - \
                           #(gauss(0, .001) * state.gaussian_rhythm * \
                           # state.tempo * 0.01)

        #### finally, a place to put the output text:
        state.outstring = state.outstring + \
             "i%1.1f %1.3f %1.3f  %s  %s  %s  %s  %s\n" % \
             (state.instr, on_time, duration, attack, pitch, pan, mix, xtra)

        ### update grid_time for next event, but only if not in chord status
        ### '[]' mode:
        if not chord_status:
            grid_time += (state.length * lf)

################################################################

def process_buffer(inbuffer, rt_mode=False):
    ''' split the whole string buffer into individual voice lines
    and feed each line to the event parser '''
    
    lines = inbuffer.splitlines()
    voiceline = 1
    current_string = '%i:' % voiceline
    ### the main loop where the input file is read:
    # the next line searches for the current instrument number
    # in the entire buffer, and only continues when there is
    # one found:
    while re.search(current_string, inbuffer):
        inst_line = ''  #### start a blank string for collection
        for line_number, line in enumerate(lines, start=1):
            text = line.rstrip() ### right strip the line
            if re.match(current_string, text):
                ### strip voice indicator prepending
                text1 = re.sub(r'^[0-9]{1,2}[:]', r'', text)
                ### get rid of barlines
                text2 = text1.replace('|', '')
                ### get rid of comments:
                text3 = re.sub(r'(.*)[#]+.*', r'\1', text2)
                inst_line += text3
                ### check syntax:
                errors = PARSER_PATTERN.split(inst_line)
                unspaced_errors = ''.join(errors).replace(' ', '')
                if unspaced_errors != '':
                    print 'there were errors at line number %i which reads:' \
                          % line_number
                    print line
                    print 'the errors are: %s' % unspaced_errors
                    print "the preprocessed line is:"
                    print inst_line
                    if not rt_mode:
                        exit()
        ## now, parse the complete voice that has been collected:
        parser(inst_line)
        ## advance the voice count and search again at the top of the loop:
        voiceline += 1
        current_string = r'%i:' % voiceline

    ### here is where we finally output the gathered results of the
    ### parser's work:
    return state.tempostring, state.outstring

##### live_loop_in function here:
def live_loop_in(instream_func=raw_input):
    ''' a function which handles interactive input. '''
    try:
        inbuff = 'i200 0 -1\n'
        while 1:
            if instream_func == raw_input:
                phrase = instream_func('microcsound--> ')
            else:
                phrase = instream_func()
            if phrase.strip() == 'done':
                return inbuff
            else:
                inbuff += phrase + '\n'
    except (KeyboardInterrupt, EOFError):
        if instream_func == raw_input:   
            print
            print "bye!"
            exit()
        ## only for testing:    
        else:
            pass
        
def sanity_tests():
    ''' runs some tests before execution... '''

    ##### test the regex pattern detection:
    test_string_1 = '''div=31 gr=0.8[L:1/4] i=17
     "2%3%.02%3%0.001%.1%.02%.5%0.001%.1%8000%.1%440%-1.2%-1.2%.001%.002"
     gf.g2.f2 p12=0.002 21 1/4 r3 @0.5 _e2 t t t t tt 4.18 [_e' _d''] '''
    try:
        assert len(PARSER_PATTERN.findall(test_string_1)) == 20
    except:
        print PARSER_PATTERN.findall(test_string_1)

    test_string_2 = '''[4.8 5.6 6.2] 15  tt 1/4 t t=80 .c4 -13 1/4 8. e3
    | e 6.12 &4 1/16 c'16 &-15 ^^f15 &-14 ^a14 &-13 ^d13 &-12 z2 ^d,, 12'''
    assert len(PARSER_PATTERN.findall(test_string_2)) == 28

    test_string_3 = "1/8 PD10 5.4 0 3 2 1 6 5 7 PU @4f 4.4 @F @5"
    assert len(PARSER_PATTERN.findall(test_string_3)) == 15
    
    ## now test the live looper function. We use a mock input via a StringIO
    ## object instead of 'raw_input', which requires human intervention.
    import StringIO
    my_test_string = "1: i=7 t=60 div=17\n1: c d e f g'\ndone\n"
    fp = StringIO.StringIO(my_test_string)
    result = live_loop_in(fp.readline)
    result_2 = process_buffer(result, rt_mode=True)
    assert "i200 0 -1\ni7.0 0.000 0.250  0.66  261.62557  0.5  1  \n" \
           "i7.0 0.250 0.250  0.66  295.66718  0.5  1  \n" \
           "i7.0 0.500 0.250  0.66  334.13815  0.5  1" \
           in result_2[1]

def main():
    ''' The place where the magic begins '''

    argparser = argparse.ArgumentParser(
                         epilog='This is microcsound v.20170228',
                         )
    argparser.usage = """microcsound [-h] [--orc orc_file] [-v] 
    [-i | 
          [[-o output_wav_file | -s, --score-only | -r, --realtime]
           [-t, --stdin | input_mc_filename ]
          ]  
    ] """
    argparser.add_argument("--orc", dest="orc_file", default=DEFAULT_ORC_FILE,
                           help="specify an orchestra file for csound to use, "\
                                " which is not the default (microcsound.orc)")
    argparser.add_argument("-v", "--debug", action="store_true",
                           dest="debug_mode",
                           help="turn on debug mode")
    argparser.add_argument("-i", "--interactive", action="store_true",
                           dest="interactive",
                           help="use an interactive prompt, render audio "\
                                "in realtime as well, does not work when "\
                                "any of -o, -s, or -r are specified")
    # outputs:                            
    argparser.add_argument("-o", "--output", dest="outwav",
                           help="optional wave file output name")
    argparser.add_argument("-s", "--score-only", action="store_true",
                           dest="score_only",
                           help="only generate a score to stdout, "\
                                " do not post-process it with csound")
    argparser.add_argument("-r", "--realtime", action="store_true", 
                           dest="realtime",
                           help="render audio in realtime")
    # inputs:
    argparser.add_argument("-t", "--stdin", action="store_true",
                           dest="text_stdin",
                           help="read text from stdin")
    argparser.add_argument("filename", nargs='?',
                           help="an input '.mc' filename or filepath")
    
    args = argparser.parse_args()

    ### if they don't know what they are doing!
    if len(argv) < 2:
        argparser.print_help()
        exit(0)
          
    ### before anything, run a set of sanity tests:
    sanity_tests()

    # wasn't argparser supposed to be helpful and do this kind of 
    # shit for us?"
    if args.interactive and (args.outwav or args.realtime \
                             or args.score_only or args.text_stdin \
                             or args.filename):
        raise argparser.error("-i must not be used with any other arguments")
    if args.outwav and (args.score_only or args.realtime):
        raise argparser.error("-o must not be used -s or -r")
    if args.score_only and (args.outwav or args.realtime):
        raise argparser.error("-s must not be used -o or -r") 
    if args.realtime and (args.outwav or args.score_only):
        raise argparser.error("-r must not be used -o or -s")     
    if (args.outwav or args.score_only or args.realtime) \
            and not (args.filename or args.text_stdin):
        raise argparser.error("You need to specify an input: a filename or "\
                              "stdin (-t)")

    # okay, we've checked all possible CL parsing errors, let's
    # figure out our working parameters:
    
    # show csound parsing messages?
    if args.debug_mode:
        verbosity_string = ''
    else:
        verbosity_string = ' -O null '
    
    if args.interactive or args.realtime:
        rt_mode = True
        csound_command = RT_CSOUND_COMMAND_STUB + verbosity_string + \
                         ' %s/%s /tmp/microcsound.sco' % (ORC_DIR, 
                                                          args.orc_file)
    else:
        rt_mode = False
        if args.filename and not args.outwav:
            out_wav = args.filename.replace('.mc', '.wav')
        else:
            out_wav = args.outwav
        csound_command = NORMAL_CSOUND_COMMAND_STUB + verbosity_string + \
                         ' -o %s %s/%s /tmp/microcsound.sco' \
                         % (out_wav, ORC_DIR, args.orc_file)

    if args.interactive:
        rt_mode = True
        while True:
            live_input = live_loop_in(instream_func=raw_input)
            print "ok...processing sound now"
            outbuf = process_buffer(live_input,
                                    rt_mode=True)
    else:
        if args.filename:
            the_file = open(args.filename)
            outbuf = process_buffer(the_file.read(),
                                    rt_mode=rt_mode)
            the_file.close()
        elif args.text_stdin:
            outbuf = process_buffer(stdin.read(),
                                    rt_mode=rt_mode)

    ## the end result, which is either a Csound score, or audio:
    if args.score_only:
        stdout.write('%s\n%s' % (outbuf[0], outbuf[1]))
    else:
        temp_sco_file = open('/tmp/microcsound.sco', 'w')
        temp_sco_file.write('%s\n%s' % (outbuf[0], outbuf[1]))
        temp_sco_file.close()
        system(csound_command)
        print

if __name__ == '__main__':
    main()
